
/*
 * events.c
 *   The libpq PGEventProc implementation.
 *
 * Copyright (c) 2008 eSilo, LLC. All rights reserved.
 * This is free software; see the source for copying conditions.  There is
 * NO warranty; not even for MERCHANTABILITY or  FITNESS FOR A  PARTICULAR
 * PURPOSE.
 */

#include "libpqtypes-int.h"

static PGtypeData *
allocTypeData(void)
{
	PGtypeData *typeData = (PGtypeData *)malloc(sizeof(PGtypeData));

	if(typeData)
		memset(typeData, 0, sizeof(PGtypeData));

	return typeData;
}

static void
freeTypeData(PGtypeData *typeData)
{
	if(typeData)
	{
		pqt_freehandlers(typeData->typhandlers, typeData->typhcnt);
		free(typeData);
	}
}

int PQtypesEventProc(PGEventId id, void *info, void *passThrough)
{
	switch(id)
	{
		case PGEVT_REGISTER:
		{
			PGEventRegister *e = (PGEventRegister *)info;
			void *data = allocTypeData();
			if(!data)
				return FALSE;
			PQsetInstanceData((PGconn *)e->conn, PQtypesEventProc, data);
			break;
		}

		case PGEVT_CONNRESET:
		{
			int i, hcnt;
			PGtypeHandler *handlers;
			PGEventConnReset *e = (PGEventConnReset *)info;
			PGtypeData *connData = (PGtypeData *)PQinstanceData(e->conn, PQtypesEventProc);

			if(!connData)
				break; /* ignore this */

			/* Have to re-register the handlers to help avoid invalidation issues.
			 * Save off the handlers and set the conn's type handlers to NULL.
			 */
			hcnt = connData->typhcnt;
			handlers = connData->typhandlers;
			connData->typhcnt = 0;
			connData->typhandlers = NULL;

			/* re-register type handlers.  It is entirely possible that a register
			 * will fail. This would almost always be because a previously registered
			 * type no longer exists in the backend (corner case).  We silently
			 * ignore these and let putf/getf alert the user ... "no such type".
			 */
			for(i=0; i < hcnt; i++)
				PQregisterTypeHandler((PGconn *)e->conn, handlers[i].orig_typname,
					handlers[i].orig_typput, handlers[i].orig_typget);

			/* free the out-dated handler list */
			pqt_freehandlers(handlers, hcnt);
			break;
		}

		case PGEVT_CONNDESTROY:
		{
			PGEventConnDestroy *e = (PGEventConnDestroy *)info;
			freeTypeData((PGtypeData *)PQinstanceData(e->conn, PQtypesEventProc));
			break;
		}

		case PGEVT_RESULTCREATE:
		{
			PGtypeData *resData;
			PGEventResultCreate *e = (PGEventResultCreate *)info;
			PGtypeData *connData = (PGtypeData *)PQinstanceData(e->conn, PQtypesEventProc);

			if(!connData || !(resData = allocTypeData()))
				return FALSE;

			/* get type formatting info from conn */
			pqt_getfmtinfo(e->conn, &resData->fmtinfo);

			/* copy type handlers from PGconn's typeData */
			if(connData->typhcnt > 0)
			{
				resData->typhandlers = pqt_duphandlers(
					connData->typhandlers, connData->typhcnt);

				if(resData->typhandlers)
					resData->typhcnt = connData->typhcnt;
			}

			PQresultSetInstanceData((PGresult *)e->result, PQtypesEventProc, resData);
			break;
		}

		case PGEVT_RESULTCOPY:
		{
			PGtypeData *destData;
			PGEventResultCopy *e = (PGEventResultCopy *)info;
			PGtypeData *srcData = (PGtypeData *)PQresultInstanceData(e->src, PQtypesEventProc);

			if(!srcData || !(destData = allocTypeData()))
				return FALSE;

			memcpy(&destData->fmtinfo, &srcData->fmtinfo, sizeof(PGtypeFormatInfo));

			/* copy type handlers from PGresult's typeData */
			if(srcData->typhcnt > 0)
			{
				destData->typhandlers = pqt_duphandlers(
					srcData->typhandlers, srcData->typhcnt);

				if(destData->typhandlers)
					destData->typhcnt = srcData->typhcnt;
			}

			PQresultSetInstanceData(e->dest, PQtypesEventProc, destData);
			break;
		}

		case PGEVT_RESULTDESTROY:
		{
			PGEventResultDestroy *e = (PGEventResultDestroy *)info;
			freeTypeData((PGtypeData *)PQresultInstanceData(e->result, PQtypesEventProc));
			break;
		}
	}

	return TRUE;
}
